#!/usr/bin/python
# -*- coding: UTF-8 -*-
#
# `Damo-Embedding` - 'c++ tool for sparse parameter server'
# Copyright (C) 2019 - present timepi <timepi123@gmail.com>
# `Damo-Embedding` is provided under: GNU Affero General Public License
# (AGPL3.0) https:#www.gnu.org/licenses/agpl-3.0.html unless stated otherwise.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#

# this file extracts sparse parts from model and updates the graph

import json
from typing import Dict, List, Tuple

import numpy as np
import torch

GLOBAL_SPARSE_EMBEDDING_USE_COUNT = 0


def is_node_for_sparse_embedding(node: torch.Node, input_node: torch.Value) -> bool:
    """is a node for a sparse embedding

    Args:
        node (torch.Node): node
        input_node (torch.Value): input node

    Returns:
        bool: is sparse embedding or not
    """
    SPARSE_TYPENAME_SUFFIX = "DummyEmbedding"
    inputs = list(node.inputs())
    if len(inputs) != 2:
        return False
    if inputs[1] != input_node:
        return False
    ntype = str(inputs[0].type())
    if not ntype.endswith(SPARSE_TYPENAME_SUFFIX):
        return False
    return True


def extract_attribute_key(node: torch.Node) -> str:
    """extract attribute key

    Args:
        node (torch.Node): sparse node to extract attribute key

    Returns:
        str: model attribute key
    """
    global GLOBAL_SPARSE_EMBEDDING_USE_COUNT
    GLOBAL_SPARSE_EMBEDDING_USE_COUNT += 1
    inputs = list(node.inputs())
    n = inputs[0].node()
    assert n.kind() == "prim::GetAttr"
    n_inputs = list(n.inputs())
    assert len(n_inputs) == 1
    assert n_inputs[0].debugName().startswith("self")
    origin = n.s("name")
    name = f"{origin}__{GLOBAL_SPARSE_EMBEDDING_USE_COUNT}"
    n.s_("name", name)
    return origin


def update_sparse_parts_from_graph(model: torch.nn.Module) -> List:
    """update graph

    Args:
        model (torch.nn.Module): model to update

    Returns:
        List: inputs process list
    """
    # get model's sparse embeddings
    embeddings = get_sparse_embeddings(model)
    inputs = list(model.graph.inputs())[1:]
    ret = []
    matches = []
    for i, node in enumerate(inputs):
        users = list(map(lambda x: x.user, node.uses()))
        if len(users) == 0:
            continue
        status = is_node_for_sparse_embedding(users[0], node)

        # if input i for sparse embedding,
        # then all nodes use this input must be sparse embedding.
        for user in users:
            s = is_node_for_sparse_embedding(user, node)
            assert s == status
        if not status:
            continue
        start = 0
        keys, dims = [], []

        for j, user in enumerate(users):
            key = extract_attribute_key(user)
            dim = embeddings[key]["dim"]
            keys.append(key)
            dims.append(dim)
            pattern, replacement = build_pattern_replacement_graph(
                user, node, start, dim
            )
            start += dim
            matches.append((pattern, replacement))
        ret.append({"input": i, "keys": keys, "dims": dims})
    for pattern, replacement in matches:
        torch._C._jit_pass_custom_pattern_based_rewrite_graph(
            pattern, replacement, model.graph
        )
    return ret


def get_sparse_embeddings(model: torch.nn.Module) -> Dict[str, Dict[str, np.ndarray]]:
    """get sparse embedding from model
    currently only supports sparse embedding as model's attributes

    Args:
        model (torch.nn.Module): torch model object

    Returns:
        Dict[str, Dict[str, np.ndarray]]: embeddings
    """
    embeddings = {}
    for k, v in model.__dict__["_modules"].items():
        if v.original_name == "DummyEmbedding":
            dim = v.embedding.weight.shape[1]
            embeddings[k] = {
                "dim": dim,
            }
    return embeddings


def build_pattern_replacement_graph(
    node: torch.Node, input_node: torch.Value, start: int, dim: int
) -> Tuple[str, str]:
    """build the replacement pattern

    Args:
        node (torch.Node): node
        input_node (torch.Value): input
        start (int): slice starting
        dim (int): slice width

    Returns:
        Tuple[str,str]: pattern and replacement
    """

    def remove_comments(code: str) -> str:
        idx = code.find("#")
        if idx >= 0:
            code = code[:idx]
        return code.strip()

    prev = node.prev()
    prev_attr = prev.s("name")
    prev_input = f"%{list(prev.inputs())[0].debugName()}"
    prev_output = f"%{list(prev.outputs())[0].debugName()}"

    input = f"%{input_node.debugName()}"
    output = f"%{list(node.outputs())[0].debugName()}"
    code_0 = f'{prev_output} = prim::GetAttr[name="{prev_attr}"]({prev_input})'
    code_1 = remove_comments(str(node))

    pattern = f"""
    graph({prev_input}, {input}): 
        {code_0}
        {code_1}
        return ({output}) 
    """
    end = start + dim
    replacement = f"""
    graph({prev_input}, {input}): 
        %1 : NoneType = prim::Constant()
        %2 : int = prim::Constant[value=-1]()
        %3 : int = prim::Constant[value=1]()
        %4 : int = prim::Constant[value={start}]()
        %5 : int = prim::Constant[value={end}]()
        %out : Tensor= aten::slice({input}, %2, %4, %5, %3) 
        return (%out)
    """
    return pattern, replacement


def update_model_graph(model: torch.nn.Module, model_path: str, meta_path: str):
    """update model graph

    Args:
        model (torch.nn.Module): original model
        model_path (str): new model path
        meta_path (str): meta file path
    """
    meta = update_sparse_parts_from_graph(model)
    json.dump({"meta": meta, "sparse": 1}, open(meta_path, "w"))
    model.save(model_path)


def update_model(original_model_path: str, model_path: str, meta_path: str):
    """update model graph

    Args:
        model (str): original model
        model_path (str): new model path
        meta_path (str): meta file path
    """
    model = torch.jit.load(original_model_path)
    update_model_graph(model=model, model_path=model_path, meta_path=meta_path)



#!/usr/bin/python
# -*- coding: UTF-8 -*-
#
# `Damo-Embedding` - 'c++ tool for sparse parameter server'
# Copyright (C) 2019 - present timepi <timepi123@gmail.com>
# `Damo-Embedding` is provided under: GNU Affero General Public License
# (AGPL3.0) https:#www.gnu.org/licenses/agpl-3.0.html unless stated otherwise.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
import json
import os
import shutil

import torch

from .util import Embedding, checkpoint


def save_model_for_training(model: torch.nn.Module, output_dir: str):
    """save model to dir, for training again

    Args:
        model (torch.nn.Module): torch module
        output_dir (str): output dir
    """
    # remove output dir
    output_dir = os.path.join(output_dir, "save_model")
    if os.path.exists(output_dir):
        shutil.rmtree(output_dir)
    os.makedirs(output_dir)
    checkpoint_dir = os.path.join(output_dir, "checkpoint")
    checkpoint(checkpoint_dir)

    def get_embedding_params(embedding: Embedding):
        return {
            "init_params": embedding.initializer,
            "opt_params": embedding.optimizer,
            "dim": embedding.dim,
            "group": embedding.group,
        }

    sparse_embedding_params = {}
    original_modules = {}
    for k, v in model.__dict__["_modules"].items():
        original_modules[k] = v
        if isinstance(v, Embedding):
            sparse_embedding_params[k] = get_embedding_params(v)
            model.__dict__["_modules"][k] = torch.nn.Embedding(1, v.dim, padding_idx=0)
        elif isinstance(v, torch.nn.ModuleList):
            modules = torch.nn.ModuleList()
            for i, m in enumerate(v):
                if isinstance(m, Embedding):
                    sparse_embedding_params[f"{k}-{i}"] = get_embedding_params(v)
                    modules.append(torch.nn.Embedding(1, v.dim, padding_idx=0))
                else:
                    modules.append(m)
            model.__dict__["_modules"][k] = modules
        elif isinstance(v, torch.nn.ModuleDict):
            modules = torch.nn.ModuleDict()
            for i, m in v.items():
                if isinstance(m, Embedding):
                    sparse_embedding_params[f"{k}-{i}"] = get_embedding_params(v)
                    modules[i] = torch.nn.Embedding(1, v.dim, padding_idx=0)
                else:
                    modules[i] = m
            model.__dict__["_modules"][k] = modules

    torch.save(model, os.path.join(output_dir, "model.pt"))
    json.dump(
        sparse_embedding_params,
        open(os.path.join(output_dir, "sparse_config.json"), "w"),
    )

    # recover
    for k, _ in model.__dict__["_modules"].items():
        model.__dict__["_modules"][k] = original_modules[k]


def load_model(dir: str) -> torch.nn.Module:
    """load model to for training again

    Args:
        dir (str): model directory

    Returns:
        torch.nn.Module: model to train
    """
    dir = os.path.join(dir, "save_model")
    sparse_embedding_params = json.load(
        open(os.path.join(dir, "sparse_config.json"), "r")
    )
    model = torch.load(os.path.join(dir, "model.pt"))

    def create_embedding(params):
        embedding = Embedding(
            dim=params["dim"],
            initializer=params["init_params"],
            optimizer=params["opt_params"],
        )
        setattr(embedding, "group", params["group"])
        return embedding

    for k, v in model.__dict__["_modules"].items():
        if isinstance(v, torch.nn.Embedding):
            key = k
            if key not in sparse_embedding_params:
                continue
            config = sparse_embedding_params[key]
            model.__dict__["_modules"][k] = create_embedding(config)
        elif isinstance(v, torch.nn.ModuleList):
            modules = torch.nn.ModuleList()
            for i, m in enumerate(v):
                key = f"{k}-{i}"
                if key not in sparse_embedding_params:
                    modules.append(m)
                else:
                    config = sparse_embedding_params[key]
                    modules.append(create_embedding(config))
            model.__dict__["_modules"][k] = modules
        elif isinstance(v, torch.nn.ModuleDict):
            modules = torch.nn.ModuleDict()
            for i, m in v.items():
                key = f"{k}-{i}"
                if key not in sparse_embedding_params:
                    modules[i] = m
                else:
                    config = sparse_embedding_params[key]
                    modules[i] = create_embedding(config)
            model.__dict__["_modules"][k] = modules
    return model



#!/usr/bin/python
# -*- coding: UTF-8 -*-
#
# `Damo-Embedding` - 'c++ tool for sparse parameter server'
# Copyright (C) 2019 - present timepi <timepi123@gmail.com>
# `Damo-Embedding` is provided under: GNU Affero General Public License
# (AGPL3.0) https:#www.gnu.org/licenses/agpl-3.0.html unless stated otherwise.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
import struct
import subprocess
import time
from collections import defaultdict
from multiprocessing import Process
from typing import List

import numpy as np
import requests
import torch

from . import config
import socket
def wait_port_ready(port,host="localhost", timeout=1200):
    start_time = time.time()
    while time.time() - start_time < timeout:
        try:
            sock = socket.create_connection((host, port), timeout=3)
            sock.close()
            cost_ts = time.time() - start_time
            print(f"wait_port_ready cost {cost_ts}")
            return True
        except (socket.timeout, ConnectionRefusedError):
            time.sleep(3)

    return False

class Embedding(torch.nn.Module):
    """embedding module for training."""

    def __init__(self, dim: int, initializer={}, optimizer={}, **kwargs):
        super(Embedding, self).__init__()
        self.dim = dim
        self.initializer = initializer
        self.optimizer = optimizer

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """embedding lookup

        Args:
            input (torch.Tensor): input values

        Returns:
            torch.Tensor: embedding values (input.shape[0], input.shape[1], self.dim)
        """

        data = input.numpy().astype(np.int64)
        batch_size, width = data.shape
        keys = np.unique(np.concatenate(data)).astype(np.int64)
        length = keys.shape[0]
        weights = np.zeros(self.dim * length, dtype=np.float32)
        pull(self.group, keys, weights)
        weights = weights.reshape((length, self.dim))
        weight_dict = {k: v for k, v in zip(keys, weights)}
        values = np.zeros(shape=(batch_size, width, self.dim), dtype=np.float32)

        for i in range(batch_size):
            for j in range(width):
                values[i][j] = weight_dict[data[i][j]]

        def apply_gradients(gradients):
            grad = gradients.numpy()
            grad = grad.reshape((batch_size, width, self.dim))
            grad_dict = defaultdict(lambda: np.zeros(self.dim, dtype=np.float32))
            for i in range(batch_size):
                for j in range(width):
                    key = data[i][j]
                    grad_dict[key] += grad[i][j]

            values = np.zeros(length * self.dim, dtype=np.float32)
            for i in range(length):
                values[i * self.dim : (i + 1) * self.dim] = (
                    grad_dict[keys[i]] / batch_size
                )

            push(self.group, keys, values)

        ret = torch.from_numpy(values)
        ret.requires_grad_()
        ret.register_hook(apply_gradients)
        return ret


def list_all_sparse_embeddings(model: torch.nn.Module) -> List[Embedding]:
    """list all sparse embeddings

    Args:
        model (torch.nn.Module): torch module

    Returns:
        List[Embedding]: all embeddings list
    """

    if isinstance(model, Embedding):
        return [model]

    embeddings = []
    for k, v in model.__dict__["_modules"].items():
        if isinstance(v, Embedding):
            embeddings.append((k, v))
        elif isinstance(v, torch.nn.ModuleList):
            for i, m in enumerate(v):
                if isinstance(m, Embedding):
                    embeddings.append((f"{k}-{i}", m))
        elif isinstance(v, torch.nn.ModuleDict):
            for i, m in v.items():
                if isinstance(m, Embedding):
                    embeddings.append((f"{k}-{i}", m))
    embeddings.sort(key=lambda x: x[0])
    return list(map(lambda x: x[1], embeddings))


def get_damo_embedding_configure(model: torch.nn.Module) -> dict:
    """get damo embedding configure for damo service or damo.PyDamo

    Args:
        model (torch.nn.Module): torch model

    Returns:
        dict: configure dict
    """
    embeddings = list_all_sparse_embeddings(model)
    configure = {"embeddings": []}
    for i, embedding in enumerate(embeddings):
        group = i
        if hasattr(embedding, "group"):
            group = getattr(embedding, "group")
        else:
            setattr(embedding, "group", i)
        configure["embeddings"].append(
            {
                "dim": embedding.dim,
                "group": group,
                "initializer": embedding.initializer,
                "optimizer": embedding.optimizer,
            }
        )
    return configure


def damo_embedding_service(config_file: str):
    """damo embedding service

    Args:
        config_file (str): configure path
    """
    subprocess.run([f"{config.DAMO_SERVICE_BINARY}", "-c", f"{config_file}"])


def run_damo_embedding_service(config_file: str, port: int):
    """run damo embedding service as daemon processor

    Args:
        config_file (str): configure path
    """
    p = Process(target=damo_embedding_service, args=(config_file,))
    p.daemon = True
    p.start()
    is_ready = wait_port_ready(port=port)
    if is_ready == False:
        raise Exception("embedding services port not ready")


def stop_damo_embeding_service():
    requests.get(f"{config.DAMO_SERVICE_ADDRESS}/stop")


def push(group: int, keys: np.ndarray, gradients: np.ndarray):
    """push gradients to damo embedding

    Args:
        group (int): embedding group
        keys (np.ndarray): push keys
        gradients (np.ndarray): gradients of keys
    """
    if config.DAMO_INSTANCE is not None:
        config.DAMO_INSTANCE.pull(group, keys, gradients)
    else:
        buffer = struct.pack(
            f"@ii{keys.shape[0]}q{gradients.shape[0]}f",
            group,
            keys.shape[0],
            *keys,
            *gradients,
        )
        headers = {
            "Content-Type": "application/octet-stream",
        }

        requests.post(f"{config.DAMO_SERVICE_ADDRESS}/push", buffer, headers=headers)


def pull(group: int, keys: np.ndarray, weights: np.ndarray):
    """pull weights from damo embedding

    Args:
        group (int): embedding group
        keys (np.ndarray): pull keys
        weights (np.ndarray): keys weights
    """
    if config.DAMO_INSTANCE is not None:
        config.DAMO_INSTANCE.pull(group, keys, weights)
    else:
        buffer = struct.pack(f"@ii{keys.shape[0]}q", group, keys.shape[0], *keys)
        headers = {
            "Content-Type": "application/octet-stream",
        }
        r = requests.post(
            f"{config.DAMO_SERVICE_ADDRESS}/pull", buffer, headers=headers
        )
        tmp = np.frombuffer(r.content, dtype=np.float32)
        assert weights.shape[0] == tmp.shape[0]
        np.copyto(weights, tmp)


def dump(dir: str):
    """dump spare embedding

    Args:
        dir (str): spare embedding path
    """
    if config.DAMO_INSTANCE is not None:
        config.DAMO_INSTANCE.dump(dir)
    else:
        headers = {
            "Content-Type": "text/plain",
        }

        requests.post(f"{config.DAMO_SERVICE_ADDRESS}/dump", data=dir, headers=headers)


def checkpoint(dir: str):
    """do checkpoint

    Args:
        dir (str): checkpoint path
    """
    if config.DAMO_INSTANCE is not None:
        config.DAMO_INSTANCE.checkpoint(dir)
    else:
        headers = {
            "Content-Type": "text/plain",
        }

        requests.post(
            f"{config.DAMO_SERVICE_ADDRESS}/checkpoint", data=dir, headers=headers
        )
